# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import Counter
from blueqat.backends.qasm_parser_backend_generator import generate_backend
import re

def draw_circuit_obaq(circuit, nqubits, depth, type=1):
  import networkx as nx
  import matplotlib.pyplot as plt

  nqubits = nqubits
  depth = depth

  if type == 1:
    width = 11
    node_color = {'x':'#04C4D9', 'y':'#F28705', 'z':'#F25C05', 'h':'#F21313', 'cx':'#04C4D9', 'cy':'#F28705', 'cz':'#F25C05', 'rx':'#04C4D9', 'ry':'#F28705', 'rz':'#F25C05'}
    node_shape = {'x':'o', 'y':'o', 'z':'o', 'h':'o', 'rx':'o', 'ry':'o', 'rz':'o'}
    node_size = 100 

  #for lines
  G = nx.Graph()

  #for init qubit
  Gi = nx.Graph()

  #for CX
  GCX = nx.Graph()
  GCY = nx.Graph()
  GCZ = nx.Graph()
  GC = nx.Graph()

  #for pauli&H
  GX = nx.Graph()
  GY = nx.Graph()
  GZ = nx.Graph()
  GH = nx.Graph()

  #for rotation
  GRX = nx.Graph()
  GRY = nx.Graph()
  GRZ = nx.Graph()

  GCarr = {'cx':GCX, 'cy':GCY, 'cz':GCZ}
  Garr = {'x':GX, 'y':GY, 'z':GZ, 'h':GH, 'rx':GRX, 'ry':GRY, 'rz':GRZ}

  #initial qubits
  Gi.add_nodes_from([(0, i) for i in range(nqubits)])

  #edge from start to final
  for i in range(nqubits):
    G.add_edge((0, i),(depth+3, i))

  #for CX
  for k, v in GCarr.items():
    for i in range(len(circuit[k])):
      GC.add_node((circuit[k][i][0], circuit[k][i][1]))
      v.add_edge((circuit[k][i][0], circuit[k][i][1]), (circuit[k][i][0], circuit[k][i][2]))

  #for gates
  for k, v in Garr.items():
    v.add_nodes_from(circuit[k])

  nx.draw(G,
        pos=dict((n, (n[0], -n[1])) for n in G.nodes()),
        node_color = 'white',
        edge_color = 'lightgrey',
        width = 1.5,
        node_size = node_size)

  nx.draw(Gi,
        pos=dict((n, (n[0], -n[1])) for n in Gi.nodes()),
        node_color = 'grey',
        node_size = node_size)

  for k, v in GCarr.items():
    nx.draw(v,
        pos=dict((n, (n[0]+1, -n[1])) for n in v.nodes()),
        node_color = node_color[k],
        edge_color = node_color[k],
        label = k.upper(),
        width = width,
        node_size = node_size)

  #controlled bit
  nx.draw(GC,
        pos=dict((n, (n[0]+1, -n[1])) for n in GC.nodes()),
        node_color = 'white',
        edge_color = 'gray',
        width = 1.5,
        node_size = 30)

  #gates
  for k, v in Garr.items():
    nx.draw(v,
        pos=dict((n, (n[0]+1, -n[1])) for n in v.nodes()),
        node_color = node_color[k],
        label = k.upper(),
        node_shape = node_shape[k],
        node_size = node_size)

  plt.legend(bbox_to_anchor=(1.15, 1),loc='upper right')
  plt.axis('equal')
  plt.show()

def _qasm_runner_obaq(qasm, shots=1024, returns=None, **kwargs):
    if returns is None:
        returns = "shots"
    elif returns not in ("shots", "draw", "_exception"):
        raise ValueError("`returns` shall be None, 'shots', 'draw', " +
                         "'_exception'")

    import_error = None
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
    except Exception as e:
        import_error = e

    if import_error:
        if returns == "_exception":
            return e
        if isinstance(import_error, ImportError):
            raise ImportError("Cannot import qiskit. To use this backend, please install networkx." +
                              " `pip install networkx `.")
        else:
            raise ValueError("Unknown error raised when importing networkx. To get exception, " +
                             'run this backend with arg `returns="_exception"`')
    else:
        if returns == "_exception":
            return None

        #parse qasm
        circ = qasm.split(';\n')
        del circ[:2]

        #nqubits
        nqubits = int(re.compile('\d+').findall(circ[0].split(' ')[1])[0])
        circ = [i for i in circ if i[:7] != "measure" and i[:4] != "qreg" and i[:4] != "creg"]

        #depth
        depth = len(circ)

        #count of gates
        circ2 = {'x':[], 'y':[], 'z':[], 'h':[], 'cx':[], 'cy':[], 'cz':[], 'rx':[], 'ry':[], 'rz':[]}
        for i in range(depth):
          temp = circ[i].split(' ')
          if temp[0] == 'cx' or temp[0] == 'cy' or temp[0] == 'cz':
            pos1 = int(re.compile('\d+').findall(temp[1].split(',')[0])[0])
            pos2 = int(re.compile('\d+').findall(temp[1].split(',')[1])[0])
            circ2[temp[0]].append([i, pos1, pos2])
          elif temp[0] == 'x' or temp[0] == 'y' or temp[0] == 'z' or temp[0] =='h':
            pos1 = int(re.compile('\d+').findall(temp[1])[0])
            circ2[temp[0]].append([i, pos1])           
          else:
            temp2 = temp[0].split('(')
            if temp2[0] == 'rx' or temp2[0] == 'ry' or temp2[0] == 'rz':
              pos1 = int(re.compile('\d+').findall(temp[1])[0])
              circ2[temp2[0]].append([i, pos1, float(temp2[1].split(')')[0])])

        #optimization of circuit
        loop = 1
        while loop > 0:
          loop = 0
          #list for check
          exist = []
          for k, v in circ2.items():
            if k == 'cx' or k == 'cy' or k == 'cz':
              for i in v:
                exist.append((i[0], i[1]))
                exist.append((i[0], i[2]))
            elif k == 'x' or k == 'y' or k == 'z' or k == 'h':
              for i in v:
                exist.append(tuple(i))
            elif k == 'rx' or k == 'ry' or k == 'rz':
              for i in v:
                exist.append((i[0], i[1]))

          #list for check
          for k, v in circ2.items():
            if k == 'cx' or k == 'cy' or k == 'cz':
              for i in range(len(v)):
                if v[i][0] != 0:
                  check1 = (v[i][0]-1, v[i][1])
                  check2 = (v[i][0]-1, v[i][2])
                  if check1 in exist or check2 in exist:
                    pass
                  else:
                    circ2[k][i] = [v[i][0]-1, v[i][1], v[i][2]]
                    loop += 1
            elif k == 'x' or k == 'y' or k == 'z' or k == 'h':
              for i in range(len(v)):
                if v[i][0] != 0:
                  check1 = (v[i][0]-1, v[i][1])
                  if check1 in exist:
                    pass
                  else:
                    circ2[k][i] = [v[i][0]-1, v[i][1]]
                    loop += 1
            elif k == 'rx' or k == 'ry' or k == 'rz':
              for i in range(len(v)):
                if v[i][0] != 0:
                  check1 = (v[i][0]-1, v[i][1])
                  if check1 in exist:
                    pass
                  else:
                    circ2[k][i] = [v[i][0]-1, v[i][1], v[i][2]]
                    loop += 1
                    
        for k, v in circ2.items():
          for i in range(len(v)):
            circ2[k][i] = tuple(circ2[k][i])

        if returns == "draw":
          return draw_circuit_obaq(circ2, nqubits, depth, type=1)

        return circ2

obaq_backend = generate_backend(_qasm_runner_obaq)

from blueqat import BlueqatGlobalSetting
BlueqatGlobalSetting.register_backend("obaq", obaq_backend)
